# Copyright (c) 2007 Ghassan Misherghi.
# Copyright (c) 2016-2017 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

import logging

from os.path import join

from .empty_dd import EmptyDD
from .unparser import Unparser

logger = logging.getLogger(__name__)


def hddmin(hdd_tree, reduce_class, reduce_config, tester_class, tester_config, test_name, work_dir,
           *, hdd_star=True, cache=None):
    """
    Run the hierarchical delta debugging reduce algorithm.

    :param hdd_tree: The root of the tree that the reduce will work with (it's the output of create_hdd_tree).
    :param reduce_class: Reference to the reducer class (LightDD, ParallelDD or CombinedParallelDD from the
                         picire module).
    :param reduce_config: Dictionary containing the parameters of the reduce_class init function.
    :param tester_class: Reference to a callable class that can decide about the interestingness of a test case.
    :param tester_config: Dictionary containing the parameters of the tester class init function (except test_builder).
    :param test_name: Name of the test case file.
    :param work_dir: Directory to save temporary test files.
    :param hdd_star: Boolean to enable the HDD star algorithm.
    :param cache: Cache to use.
    :return: The 1-tree-minimal test case.
    """

    def collect_level_ids(level):
        # Don't return a set from collect_level_ids straight away.
        # Node IDs are python object IDs, which are implementation specific
        # (memory addresses in CPython). Memory addresses may/do change from
        # execution to execution, so would change the IDs, and so may change
        # how the ID set is serialized into an ID list (necessary to construct
        # the input config for ddmin). But then, ddmin could potentially give
        # different results -- for the same tree.
        # If a list is used to collect the node IDs, however, then they will
        # always be in the same order, no matter what the actual IDs are.
        def _collect_level_ids(node, current_level):
            if current_level == level and node.state == node.KEEP:
                level_ids.append(node.id)
            return current_level + 1
        level_ids = []
        hdd_tree.inherited_attribute(_collect_level_ids, 0)
        return level_ids

    iter_cnt = 0

    while True:
        logger.info('Iteration #%d', iter_cnt)
        hdd_tree.check()

        level = 0
        changed = False
        level_ids = collect_level_ids(level)

        while len(level_ids):
            logger.info('Checking level %d ...', level)

            level_ids_set = set(level_ids)

            test_builder = Unparser(hdd_tree, level_ids_set)
            if hasattr(cache, 'set_test_builder'):
                cache.set_test_builder(test_builder)

            test = tester_class(test_builder=test_builder,
                                test_pattern=join(work_dir, 'iter_%d' % iter_cnt, 'level_%d' % level, '%s', test_name),
                                **tester_config)
            dd = reduce_class(test, cache=cache, **reduce_config)
            c = dd.ddmin(level_ids)
            if len(c) == 1:
                dd = EmptyDD(test, cache=cache)
                c = dd.ddmin(c)
            c = set(c)
            changed = changed or len(c) < len(level_ids_set)
            if cache:
                cache.clear()

            hdd_tree.set_state(level_ids_set, c)

            level += 1
            level_ids = collect_level_ids(level)

        if not hdd_star or not changed:
            break

        iter_cnt += 1

    return hdd_tree.unparse()



# Copyright (c) 2016-2017 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

import json
import logging
import re
import sys

from antlr4 import *
from antlr4.Token import CommonToken
from pkgutil import get_data
from os import makedirs, pathsep
from os.path import basename, join
from string import Template
from subprocess import CalledProcessError, Popen, PIPE

from .grammar_analyzer import analyze_grammars
from .parser_builder import build_grammars
from ..hdd_tree import HDDRule, HDDToken, HDDTree


logger = logging.getLogger(__name__)


class HDDQuantifier(HDDRule):
    """
    Special rule type in the HDD tree to support optional quantifiers.
    """
    def __init__(self, *, start=None, end=None):
        HDDRule.__init__(self, '', start=start, end=end)


class HDDErrorToken(HDDToken):
    """
    Special token type that represents unmatched tokens. The minimal replacement of such nodes
    is an empty string.
    """
    def __init__(self, text, *, start, end):
        HDDToken.__init__(self, '', text, start=start, end=end)


# Override ConsoleErrorListener to suppress parse issues in non-verbose mode.
class ConsoleListener(error.ErrorListener.ConsoleErrorListener):
    def syntaxError(self, recognizer, offendingSymbol, line, column, msg, e):
        logger.debug('line %d:%d %s', line, column, msg)
error.ErrorListener.ConsoleErrorListener.INSTANCE = ConsoleListener()


def create_hdd_tree(input_stream, grammar, start_rule, antlr, work_dir, *, replacements=None, island_desc=None, lang='python'):
    """
    Build a tree that the HDD algorithm can work with.

    :param input_stream: ANTLR stream (FileStream or InputStream) representing the input.
    :param grammar: List of the grammars describing the input format.
    :param start_rule: The name of the start rule of the parser.
    :param antlr: Path to the ANTLR4 tool (Java jar binary).
    :param work_dir: Working directory.
    :param replacements: Dictionary containing the minimal replacements of the target grammar's rules.
    :param island_desc: List of IslandDescriptor objects.
    :param lang: The target language of the parser.
    :return: The root of the created HDD tree.
    """

    grammar_workdir = join(work_dir, 'target')

    def inject_optional_actions(grammar, positions, target_file):
        """
        Update the original parser grammar by injecting actions to the start and
        end of every quantified part.

        :param grammar: Path to the grammar to be updated.
        :param positions: Start and end locations of quantified elements.
        :param target_file: Path to the updated grammar.
        """
        with open(grammar, 'rb') as f:
            lines = f.read().splitlines(keepends=True)

        languages = {
            'python': {
                'prefix': b'({self.enter_optional()} ',
                'postfix': b' {self.exit_optional()})'
            },
            'java': {
                'prefix': b'({ try { getClass().getMethod("enter_optional").invoke(this); } catch (Exception e) { assert false; }} ',
                'postfix': b' { try { getClass().getMethod("exit_optional").invoke(this); } catch (Exception e) { assert false; }})'
            }
        }

        for ln in positions:
            offset = 0
            for i, position in enumerate(sorted(positions[ln], key=lambda x: x[1])):
                if position[0] == 's':
                    lines[ln - 1] = lines[ln - 1][0:position[1] + offset] + languages[lang]['prefix'] + lines[ln - 1][position[1] + offset:]
                    offset += len(languages[lang]['prefix'])
                elif position[0] == 'e':
                    lines[ln - 1] = lines[ln - 1][0:position[1] + offset] + languages[lang]['postfix'] + lines[ln - 1][position[1] + offset:]
                    offset += len(languages[lang]['postfix'])

        with open(target_file, 'wb') as f:
            f.write(b''.join(lines))

    def build_antlr_grammars():
        """
        Build parsers to parse ANTLR4 grammars.

        :return: References to the ANTLR4 lexer and parser classes.
        """
        antlr4_workdir = join(work_dir, 'antlr4')
        makedirs(antlr4_workdir, exist_ok=True)
        if antlr4_workdir not in sys.path:
            sys.path.append(antlr4_workdir)

        # Copy the resources needed to interpret the input grammars.
        for resource in ['LexerAdaptor.py', 'ANTLRv4Lexer.g4', 'ANTLRv4Parser.g4', 'LexBasic.g4']:
            with open(join(antlr4_workdir, resource), 'wb') as f:
                f.write(get_data(__package__, join('resources', resource)))

        antlr_lexer_class, antlr_parser_class, _ = build_grammars((join(antlr4_workdir, 'ANTLRv4Lexer.g4'),
                                                                   join(antlr4_workdir, 'ANTLRv4Parser.g4'),
                                                                   join(antlr4_workdir, 'LexBasic.g4')),
                                                                  antlr4_workdir,
                                                                  antlr)
        logger.debug('ANTLR4 grammars processed...')
        return antlr_lexer_class, antlr_parser_class

    def java_classpath():
        return pathsep.join([antlr, grammar_workdir])

    def compile_java_sources(lexer, parser, listener):
        executor = Template(get_data(__package__, join('resources', 'ExtendedTargetParser.java')).decode('utf-8'))
        with open(join(grammar_workdir, 'Extended{parser}.java'.format(parser=parser)), 'w') as f:
            f.write(executor.substitute(dict(lexer_class=lexer,
                                             parser_class=parser,
                                             listener_class=listener)))
        cmd = 'javac -classpath {classpath} *.java'.format(classpath=java_classpath())
        with Popen(cmd, shell=True, cwd=grammar_workdir, stdout=PIPE, stderr=PIPE) as proc:
            stdout, stderr = proc.communicate()
            if proc.returncode != 0:
                logger.error('Java compile failed!\n%s\n%s\n', stdout, stderr)
                raise CalledProcessError(returncode=proc.returncode, cmd=cmd, output=stdout + stderr)

    def island_desc_to_list(island_desc):
        island_desc = island_desc if island_desc else []
        return island_desc if isinstance(island_desc, list) else [island_desc]

    def prepare_parsing(grammar, *, replacements=None, island_desc=None):
        """
        Performs initiative steps needed to parse the input test case (like create directory structures,
        builds grammars, sets PATH, etc...)

        :param grammar: List of the grammars describing the input format.
        :param replacements: Dictionary containing the minimal replacements of the target grammar's rules.
        :param island_desc: List of IslandDescriptor objects.
        :return: Tuple of lexer, parser, listener class references and the replacement dictionary.
        """
        antlr_lexer_class, antlr_parser_class = build_antlr_grammars()
        replacements, action_positions = analyze_grammars(antlr_lexer_class, antlr_parser_class, grammar, replacements)
        logger.debug('Replacements are calculated...')

        makedirs(grammar_workdir, exist_ok=True)
        if grammar_workdir not in sys.path:
            sys.path.append(grammar_workdir)

        # Inject actions into the target grammars to help localizing part of the test case that are optional.
        for i, g in enumerate(grammar):
            grammar[i] = join(grammar_workdir, basename(g))
            inject_optional_actions(g, action_positions[g], grammar[i])

        target_lexer_class, target_parser_class, target_listener_class = build_grammars(tuple(grammar), grammar_workdir, antlr, lang)
        logger.debug('Target grammars are processed...')

        if lang == 'java':
            compile_java_sources(target_lexer_class, target_parser_class, target_listener_class)
            return target_lexer_class, target_parser_class, target_listener_class, replacements

        island_rules = [desc.rule for desc in island_desc_to_list(island_desc)]

        class ExtendedTargetParser(target_parser_class):
            """
            ExtendedTargetParser is a subclass of the original parser implementation.
            It can trigger state changes that are needed to identify parts of the input
            that are not needed to keep it syntactically correct.
            """
            def enter_optional(self):
                self.trigger_listener('enter_optional')

            def exit_optional(self):
                self.trigger_listener('exit_optional')

            def enterRecursionRule(self, localctx: ParserRuleContext, state: int, ruleIndex: int, precedence: int):
                target_parser_class.enterRecursionRule(self, localctx, state, ruleIndex, precedence)
                self.trigger_listener('recursion_enter')

            def pushNewRecursionContext(self, localctx: ParserRuleContext, state: int, ruleIndex: int):
                target_parser_class.pushNewRecursionContext(self, localctx, state, ruleIndex)
                self.trigger_listener('recursion_push')

            def unrollRecursionContexts(self, parentCtx: ParserRuleContext):
                target_parser_class.unrollRecursionContexts(self, parentCtx)
                self.trigger_listener('recursion_unroll')

            def trigger_listener(self, event):
                for listener in self.getParseListeners():
                    if hasattr(listener, event):
                        getattr(listener, event)()

            def syntax_error_warning(self):
                if self._syntaxErrors:
                    logger.warning('%s finished with %d syntax errors. This may decrease reduce quality.',
                                   target_parser_class.__name__, self._syntaxErrors)

        class ExtendedTargetListener(target_listener_class):
            """
            ExtendedTargetListener is a subclass of the original listener implementation.
            It can trigger state changes that are needed to identify parts of the input
            that are not needed to keep it syntactically correct.
            """
            def __init__(self, parser):
                self.parser = parser
                self.current_node = None
                self.island_nodes = []
                self.root = None

            def recursion_enter(self):
                assert isinstance(self.current_node, HDDRule)
                node = HDDRule(self.current_node.name)
                self.current_node.add_child(node)
                self.current_node.recursive_rule = True
                self.current_node = node

            def recursion_push(self):
                assert len(self.current_node.parent.children) > 0

                first_child = self.current_node.parent.children[0]
                self.current_node.parent.remove_child(first_child)
                self.current_node.add_child(first_child)

            def recursion_unroll(self):
                assert self.current_node.recursive_rule
                assert len(self.current_node.children) == 1 and self.current_node.name == self.current_node.children[0].name
                children_to_lift = self.current_node.children[0].children
                parent = self.current_node.parent
                if children_to_lift:
                    self.current_node.children = []
                    self.current_node.add_children(children_to_lift)
                    self.current_node.start = self.current_node.children[0].start
                    self.current_node.end = self.current_node.children[-1].end
                else:
                    parent.remove_child(self.current_node)
                self.current_node = parent

            def enterEveryRule(self, ctx:ParserRuleContext):
                name = self.parser.ruleNames[ctx.getRuleIndex()]
                node = HDDRule(name)
                if not self.root:
                    self.root = node
                else:
                    assert self.current_node
                    self.current_node.add_child(node)
                self.current_node = node

            def exitEveryRule(self, ctx:ParserRuleContext):
                # If the input contains syntax error, then the last optional block was may not closed.
                while isinstance(self.current_node, HDDQuantifier):
                    self.exit_optional()

                assert self.current_node.name == self.parser.ruleNames[ctx.getRuleIndex()],\
                    '%s (%s) != %s' % (self.current_node.name, repr(self.current_node), self.parser.ruleNames[ctx.getRuleIndex()])

                start, _ = self.tokenBoundaries(ctx.start)
                _, end = self.tokenBoundaries(ctx.stop if ctx.stop else ctx.start)
                self.current_node.start = start
                self.current_node.end = end

                if self.current_node.parent:
                    self.current_node = self.current_node.parent

            def tokenBoundaries(self, token):
                line_breaks = token.text.count('\n')
                return HDDTree.Position(token.line, token.column), \
                       HDDTree.Position(token.line + line_breaks,
                                        token.column + len(token.text) if not line_breaks else
                                        len(token.text) - token.text.rfind('\n'))

            def visitTerminal(self, ctx:TerminalNode):
                name, text = (self.parser.symbolicNames[ctx.symbol.type], ctx.symbol.text) if ctx.symbol.type != Token.EOF else ('EOF', '')
                start, end = self.tokenBoundaries(ctx.symbol)

                node = HDDToken(name,
                                text,
                                start=start,
                                end=end)
                self.current_node.add_child(node)
                if name in island_rules:
                    self.island_nodes.append(node)

            def visitErrorNode(self, ctx:ErrorNode):
                if hasattr(ctx, 'symbol'):
                    start, end = self.tokenBoundaries(ctx.symbol)
                    self.current_node.add_child(HDDErrorToken(ctx.symbol.text, start=start, end=end))

            def enter_optional(self):
                quant_node = HDDQuantifier()
                self.current_node.add_child(quant_node)
                self.current_node = quant_node

            def exit_optional(self):
                assert self.current_node.parent, 'Quantifier node has no parent.'
                assert self.current_node.children, 'Quantifier node has no children.'

                self.current_node.start = self.current_node.children[0].start
                self.current_node.end = self.current_node.children[-1].end
                self.current_node = self.current_node.parent

            def print_tree(self):
                if self.root and logger.isEnabledFor(logging.DEBUG):
                    logger.debug(self.root.tree_str(current=self.current_node))

        return target_lexer_class, ExtendedTargetParser, ExtendedTargetListener, replacements

    class ExtendedErrorListener(error.ErrorListener.ErrorListener):

        def syntaxError(self, recognizer, offendingSymbol, line, column, msg, e):
            t = CommonToken(source=(recognizer, recognizer._input),
                            type=Token.INVALID_TYPE,
                            channel=Token.DEFAULT_CHANNEL,
                            start=recognizer._tokenStartCharIndex,
                            stop=recognizer._tokenStartCharIndex)
            t.line = recognizer._tokenStartLine
            t.column = recognizer._tokenStartColumn
            recognizer._type = Token.MIN_USER_TOKEN_TYPE
            recognizer.emitToken(t)

    def build_hdd_tree(input_stream, lexer_class, parser_class, listener_class, start_rule, island_desc, replacements):
        """
        Parse the input with the provided ANTLR classes.

        :param input_stream: ANTLR stream (FileStream or InputStream) representing the input.
        :param lexer_class: Reference to the lexer class.
        :param parser_class: Reference to the parser class.
        :param listener_class: Reference to the listener class.
        :param start_rule: The name of the start rule of the parser.
        :param island_desc: List of IslandDescriptor objects.
        :param replacements: Dictionary containing the minimal replacements of the target grammar's rules.
        :return: The root of the created HDD tree.
        """

        def set_replacement(node):
            if isinstance(node, (HDDQuantifier, HDDErrorToken)):
                node.replace = ''
            elif isinstance(node, HDDRule):
                node.replace = replacements[node.name]
            else:
                node.replace = replacements.get(node.name, node.text)

        island_nodes = []

        logger.debug('Parse input with %s rule', start_rule)
        if lang != 'python':
            island_rules = [desc.rule for desc in island_desc_to_list(island_desc)]

            def hdd_tree_from_json(node_dict):
                # Convert interval dictionaries to Position objects.
                node_dict.update({
                    'start': HDDTree.Position(**node_dict['start']),
                    'end': HDDTree.Position(**node_dict['end'])})

                name = node_dict.get('name', None)
                children = node_dict.pop('children', None)
                cls = eval(node_dict.pop('type'))
                node = cls(**node_dict)

                if children:
                    for child in children:
                        node.add_child(hdd_tree_from_json(child))
                elif name:
                    if name in island_rules:
                        island_nodes.append(node)
                return node

            cmd = 'java -classpath {classpath} Extended{parser} {start_rule}'.format(classpath=java_classpath(),
                                                                                     parser=parser_class,
                                                                                     start_rule=start_rule)
            with Popen(cmd,
                       cwd=grammar_workdir,
                       shell=True,
                       stdin=PIPE,
                       stdout=PIPE,
                       stderr=PIPE,
                       universal_newlines=True) as proc:
                stdout, stderr = proc.communicate(input=input_stream.strdata)
                if proc.returncode != 0:
                    logger.error('Java parser failed!\n%s\n%s\n', stdout, stderr)
                    raise CalledProcessError(returncode=proc.returncode, cmd=cmd, output=stdout + stderr)
            tree_root = hdd_tree_from_json(json.loads(stdout))
        else:
            lexer = lexer_class(input_stream)
            lexer.addErrorListener(ExtendedErrorListener())
            target_parser = parser_class(CommonTokenStream(lexer))
            parser_listener = listener_class(target_parser)
            target_parser.addParseListener(parser_listener)

            getattr(target_parser, start_rule)()
            target_parser.syntax_error_warning()
            island_nodes = parser_listener.island_nodes
            assert parser_listener.root == parser_listener.current_node
            tree_root = parser_listener.root

        # Traverse the HDD tree and set minimal replacements for nodes.
        tree_root.traverse(set_replacement)
        process_island_nodes(island_nodes, island_desc)
        logger.debug('Parse done.')
        return tree_root

    def process_island_nodes(island_nodes, island_desc):
        islands = dict()
        island_desc = island_desc_to_list(island_desc)

        for node in island_nodes:
            if node.name not in islands:
                island = next(filter(lambda island: island.rule == node.name, island_desc))
                lexer, parser, listener, replacements = prepare_parsing(grammar=island.grammars,
                                                                        replacements=island.replacements,
                                                                        island_desc=island.island_desc)
                islands[node.name] = (island.pattern, island.island_desc, lexer, parser, listener, replacements)

            new_node = HDDRule(node.name, start=node.start, end=node.end)
            new_node.add_children(build_island_subtree(node, *islands[node.name]))
            node.replace_with(new_node)

    def build_island_subtree(node, pattern, desc, lexer, parser, listener, replacements):
        """
        Process terminal with an island grammar.

        :param node: HDDToken object containing island language.
        :return: List of HDDTree nodes representing the `children` of node.
        """
        last_processed = 0
        content = node.text
        children = []

        # Intervals describes a non-overlapping splitting of the content according to the pattern.
        intervals = []
        for m in re.finditer(pattern, content):
            intervals.extend([(g, m.start(g), m.end(g)) for g in list(pattern.groupindex.keys())])
        intervals.sort(key=lambda x: (x[1], x[2]))

        for interval in intervals:
            # Create simple HDDToken of the substring proceeding a subgroup.
            if last_processed < interval[1]:
                next_token_text = content[last_processed:interval[1]]
                prefix = content[0:last_processed]
                children.append(HDDToken(name='',
                                         text=next_token_text,
                                         start=HDDTree.Position(node.start.line + content[0:last_processed].count('\n'),
                                                                len(prefix) - prefix.rfind('\n')),
                                         end=HDDTree.Position(node.start.line + next_token_text.count('\n'),
                                                              len(next_token_text) - next_token_text.rfind('\n')),
                                         replace=next_token_text))

            # Process a island and save its subtree.
            children.append(build_hdd_tree(input_stream=InputStream(content[interval[1]:interval[2]]),
                                           lexer_class=lexer,
                                           parser_class=parser,
                                           listener_class=listener,
                                           start_rule=interval[0],
                                           island_desc=desc,
                                           replacements=replacements))
            last_processed = interval[2]

        # Create simple HDDToken of the substring following the last subgroup if any.
        if last_processed < len(content):
            next_token_text = content[last_processed:]
            prefix = content[0:last_processed]
            children.append(HDDToken(name='',
                                     text=next_token_text,
                                     start=HDDTree.Position(node.start.line + content[0:last_processed].count('\n'),
                                                            len(prefix) - prefix.rfind('\n')),
                                     end=HDDTree.Position(node.start.line + next_token_text.count('\n'),
                                                          len(next_token_text) - next_token_text.rfind('\n')),
                                     replace=next_token_text))
        return children

    def remove_empty_children(node):
        if isinstance(node, HDDRule):
            non_empty_children = []

            for child in node.children:
                if isinstance(child, HDDToken):
                    # empty token is usually the EOF only (but interestingly, it may
                    # appear multiple times in the tree)
                    if child.text != '':
                        non_empty_children.append(child)
                else:
                    assert isinstance(child, HDDRule)
                    remove_empty_children(child)

                    # a grammar may contain lambda rules (with nothing on the
                    # right-hand side, or with an empty alternative), or rules that
                    # produce EOF only (which is removed in the branch above)
                    if len(child.children) != 0:
                        non_empty_children.append(child)

            node.children[:] = non_empty_children

    lexer_class, parser_class, listener_class, replacements = prepare_parsing(grammar=grammar,
                                                                              replacements=replacements,
                                                                              island_desc=island_desc)
    tree = build_hdd_tree(input_stream=input_stream,
                          lexer_class=lexer_class,
                          parser_class=parser_class,
                          listener_class=listener_class,
                          start_rule=start_rule,
                          island_desc=island_desc,
                          replacements=replacements)
    remove_empty_children(tree)
    return tree



# Copyright (c) 2016-2017 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

from .cli import __version__, call
from .coarse_hdd import coarse_hddmin, coarse_full_hddmin
from .hdd import hddmin
from .hdd_tree import HDDRule, HDDToken, HDDTree

__all__ = [
    '__version__',
    'call',
    'cli',
    'coarse_full_hddmin',
    'coarse_hddmin',
    'hddmin',
    'HDDRule',
    'HDDToken',
    'HDDTree',
]



# Copyright (c) 2007 Ghassan Misherghi.
# Copyright (c) 2016-2017 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.


class HDDTree:
    # Node states for unparsing.
    REMOVED = 0
    KEEP = 1

    def __init__(self, name, *, start=None, end=None, replace=None):
        """
        Initialize a HDD tree/node.

        :param name: The name of the node.
        :param start: Position object describing the start of the HDDTree node.
        :param end: Position object describing the end of the HDDTree node.
        :param replace: The minimal replacement string of the current node.
        """
        self.name = name
        self.replace = replace
        self.start = start
        self.end = end
        self.parent = None
        self.state = self.KEEP
        self.id = id(self)

    class Position(object):
        """Class defining a position in the input file. Used to recognise line breaks between tokens."""
        def __init__(self, line, idx):
            """
            Initialize position object.

            :param line: Line index in the input.
            :param idx: Absolute character index in the input.
            """
            self.line = line
            self.idx = idx

    def traverse(self, visitor):
        """
        Function providing depth-first traversal for a visitor function.

        :param visitor: Function applying to the visited nodes.
        """
        assert False, 'Should never be reached: it should be overridden in sub-classes.'

    def synthetic_attribute(self, visitor):
        """
        Call visitor on nodes without propagating any values (used by unparsing).

        :param visitor: Function applying on the visited nodes.
        :return: The value returned by visitor after applying on the node.
        """
        assert False, 'Should never be reached: it should be overridden in sub-classes.'

    def inherited_attribute(self, visitor, attribute=None):
        """
        Call visitor on the nodes and propagate the return value to the children (only setLevel uses it).

        :param visitor: Function applying to the visited nodes.
        :param attribute: The propagated value.
        """
        assert False, 'Should never be reached: it should be overridden in sub-classes.'

    def unparse(self):
        """
        Build test case from a HDD tree.

        :return: The unparsed test case.
        """
        def unparse_attribute(node, attribs):
            if node.state != self.KEEP:
                return node.replace

            # Keep the text of the token.
            if isinstance(node, HDDToken):
                return node.text

            if not attribs:
                return ''

            # Concat the text of children.
            assert node.children
            test_src = attribs[0]
            if len(node.children) > 1:
                for i in range(1, len(node.children)):
                    # Do not add extra spaces if the next chunk is empty.
                    if not attribs[i]:
                        continue
                    if node.children[i].start.line > node.children[i - 1].end.line:
                        test_src += '\n'
                    elif node.children[i].start.idx > node.children[i - 1].end.idx:
                        test_src += ' '
                    test_src += attribs[i]

            return test_src

        return self.synthetic_attribute(unparse_attribute)

    def set_state(self, ids, keepers):
        """
        Set the status of some selected nodes: if they are in the collection of
        keepers, mark them as kept, otherwise mark them as removed.

        :param ids: The collection (list or set) of node IDs to set state for.
        :param keepers: The collection (list or set) of IDs to be kept.
        """
        def _set_state(node):
            if node.id in ids:
                node.state = self.KEEP if node.id in keepers else self.REMOVED
        self.traverse(_set_state)

    def check(self):
        """Run sanity check on the HDD tree."""
        def bad_parent(node):
            if node is None:
                return
            assert isinstance(node, HDDToken) or None not in node.children, 'Bad parent node: %s' % node.name
        self.traverse(bad_parent)

    def tree_str(self, *, current=None):
        """
        Pretty print HDD tree to help debugging.

        :param current: Reference to a node that will be marked with a '*' in the output.
        :return: String representation of the tree.
        """

        def _tree_str(node, attrib):
            if node.state != node.KEEP:
                return ''

            return '[%s:%s]%s%s%s(%s)\n%s' % (
                node.name,
                node.__class__.__name__,
                ('"%s"' % node.text) if isinstance(node, HDDToken) else '',
                ('(ln:%d,i:%d)-(ln:%d,i:%d)' % (node.start.line, node.start.idx, node.end.line,
                                                node.end.idx)) if self.start is not None and self.end is not None else '',
                '*' if node == current else '',
                node.replace,
                ''.join(['    ' + line + '\n' for line in ''.join(attrib).splitlines()]))

        return self.synthetic_attribute(_tree_str)

    def squeeze_tree(self):
        """
        Suppress single line chains in the HDD tree whose minimal replacements are the
        same and hence they would result in redundant checks during the minimization.
        """
        return self

    def skip_unremovable_tokens(self):
        """
        Mark those tokens as removed whose text is the same as their minimal replacement,
        thus hiding them from hddmin, because they just cause extra test runs but cannot reduce the input.
        """
        pass

    def replace_with(self, other):
        """
        Replace the current node with `other` in the HDD tree.

        :param other: Node to replace the current with.
        """
        self.parent.children[self.parent.children.index(self)] = other
        other.parent = self.parent

    def flatten_recursion(self):
        """
        Heuristics to flatten left or right-recursion. E.g., given a rule
            rule : a | rule b
        and a HDD tree built with it from an input, rewrite the resulting HDD
        tree as if it was built using
            rule : a b*
        This allows HDD to potentially completely remove the recurring blocks
        (instead of replacing them with their minimal replacement, which is
        usually not "").
        """
        if isinstance(self, HDDRule) and self.state == self.KEEP:
            for child in self.children:
                child.flatten_recursion()

            if len(self.children) > 1:
                if self.children[0].name == self.name:
                    left = self.children[0]

                    right = HDDRule('', replace='', start=self.children[1].start, end=self.children[-1].end)
                    right.add_children(self.children[1:])
                    del self.children[:]

                    self.add_children(left.children)
                    self.add_child(right)

                elif self.children[-1].name == self.name:
                    right = self.children[-1]

                    left = HDDRule('', replace='', start=self.children[0].start, end=self.children[-2].end)
                    left.add_children(self.children[0:-1])
                    del self.children[:]

                    self.add_child(left)
                    self.add_children(right.children)


class HDDToken(HDDTree):
    def __init__(self, name, text, *, start, end, replace=None):
        HDDTree.__init__(self, name, start=start, end=end, replace=replace)
        self.text = text

    def traverse(self, visitor):
        visitor(self)

    def synthetic_attribute(self, visitor):
        return visitor(self, [])

    def inherited_attribute(self, visitor, attribute=None):
        visitor(self, attribute)

    def skip_unremovable_tokens(self):
        if self.text == self.replace:
            self.state = self.REMOVED


class HDDRule(HDDTree):
    def __init__(self, name, *, start=None, end=None, replace=None):
        HDDTree.__init__(self, name, start=start, end=end, replace=replace)
        self.children = []

    def add_child(self, child):
        self.children.append(child)
        child.parent = self

    def add_children(self, children):
        for child in children:
            self.add_child(child)

    def remove_child(self, child):
        self.children.remove(child)

    def traverse(self, visitor):
        visitor(self)
        if self.state != self.KEEP:
            return
        for child in self.children:
            child.traverse(visitor)

    def synthetic_attribute(self, visitor):
        if self.state != self.KEEP:
            return visitor(self, [])
        return visitor(self, [child.synthetic_attribute(visitor) for child in self.children])

    def inherited_attribute(self, visitor, attribute=None):
        inherit_value = visitor(self, attribute)

        if self.state != self.KEEP:
            return

        for child in self.children:
            child.inherited_attribute(visitor, inherit_value)

    def squeeze_tree(self):
        for i, child in enumerate(self.children):
            self.children[i].replace_with(child.squeeze_tree())

        if len(self.children) == 1 and self.children[0].replace == self.replace:
            return self.children[0]

        return self

    def skip_unremovable_tokens(self):
        for child in self.children:
            child.skip_unremovable_tokens()



# Copyright (c) 2017 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

import logging

from picire import AbstractDD, ConfigCache

logger = logging.getLogger(__name__)


class EmptyDD(AbstractDD):
    """
    Special DD variant that *does* test the empty configuration (and nothing
    else).
    """

    def __init__(self, test, *, cache=None):
        """
        Initialize an EmptyDD object.

        :param test: A callable tester object.
        :param cache: Cache object to use.
        """
        AbstractDD.__init__(self, test, None, cache=cache)

    def ddmin(self, config):
        """
        Return a 1-minimal failing subset of the initial configuration, and also
        test the empty configuration while doing so.

        Note: The initial configuration is expected to be of size 1, thus the
        1-minimal failing subset is always its trivial subset: either itself or
        the empty configuration.

        :param config: The initial configuration that will be reduced.
        :return: 1-minimal failing configuration.
        """

        assert len(config) == 1

        logger.debug('dd(%r) ...', config)

        outcome = self._dd(config)

        logger.debug('dd(%r) = %r', config, outcome)

        return outcome

    def _dd(self, config):
        # assert self.test(config, 'assert') == self.FAIL

        emptyset = []
        config_id = 'empty'

        logger.info('Run: trying 0.')

        outcome = self.lookup_cache(emptyset, config_id) or self.test(emptyset, config_id)
        if outcome == self.FAIL:
            logger.info('Reduced to 0 units.')
            logger.debug('New config: %r.', emptyset)

            logger.info('Done.')
            return emptyset

        logger.info('Done.')
        return config



# Copyright (c) 2016 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

import logging
import re

logger = logging.getLogger(__name__)


# Paser Elements

class ANTLRElement(object):
    def __init__(self, *, optional=False, repl=None):
        """
        Constructor of the base tree node type.

        :param optional: Boolean indicating whether the current node is optional or not.
        :param repl: Known replacement if any.
        """
        self.children = []
        self.replacement = repl if not optional else ''

    def all_replacements_defined(self):
        """
        Replacements are defined if the node has at least one child and all of the children
        have a replacement set.
        """
        return self.children and all(x.replacement is not None for x in self.children)

    def has_defined_replacement(self):
        """
        Checks if any of the children has a defined replacement.
        Needed by alternations since the replacement of a recursive rule wouldn't be possible to
        determine if waiting for all the children set.
        """
        return self.children and any(x.replacement is not None for x in self.children)

    def calc_replacement(self):
        """
        The minimal replacement of a parser rule is the concatenation of its children's minimal replacement.

        :return: Boolean denoting if a new replacement was found or not.
        """
        if self.all_replacements_defined():
            new_repl = ''.join([x.replacement for x in self.children])
            if self.replacement is None or len(new_repl) < len(self.replacement) or (len(new_repl) == len(self.replacement) and new_repl < self.replacement):
                self.replacement = new_repl
                return True
        return False


class ANTLRRule(ANTLRElement):
    """
    Representation of a parser rule. The replacement string determined here
    will be used in the reduce phase. This replacement can be set by the user
    or generated automatically. If it's set by the user then it won't be changed
    ever (even if it isn't minimal).
    """
    def __init__(self, name, repl):
        ANTLRElement.__init__(self, repl=repl)
        self.name = name
        self.const_replacement = repl is not None

    def calc_replacement(self):
        if self.const_replacement:
            return False
        return ANTLRElement.calc_replacement(self)


class ANTLRRef(ANTLRElement):
    def __init__(self, ref, *, optional=False):
        ANTLRElement.__init__(self, optional=optional)
        self.ref = ref


class ANTLRAlternation(ANTLRElement):
    def calc_replacement(self):
        """
        The minimal replacement of an alternation is it's shortest child.

        :return: Boolean denoting if a new replacement was found or not.
        """
        if self.has_defined_replacement():
            new_repl = min(list(filter(lambda i: i is not None, [c.replacement for c in self.children])), key=len, default=False)
            if self.replacement is None or len(new_repl) < len(self.replacement) or (len(new_repl) == len(self.replacement) and new_repl < self.replacement):
                self.replacement = new_repl
                return True
        return False


# Lexer Elements

class ANTLRLexerElement(ANTLRElement):
    def __init__(self, *, optional=False, repl=None):
        ANTLRElement.__init__(self, optional=optional, repl=repl)
        self.start_intervals = None

    def starters_defined(self):
        return self.children and all(x.start_intervals is not None for x in self.children)

    def calc_starters(self):
        if self.start_intervals is None and self.starters_defined():
            self.start_intervals = sum((x.start_intervals for x in self.children), [])
            return True
        return False


class ANTLRLexerRule(ANTLRLexerElement):
    """
    Representation of a lexer rule. The replacement string determined here
    will be used in the reduce phase. This replacement can be set by the user
    or generated automatically. If it's set by the user then it won't be changed
    ever (even if it's not minimal).
    """
    def __init__(self, name, repl):
        ANTLRLexerElement.__init__(self, repl=repl)
        self.name = name
        self.const_replacement = repl is not None

    def calc_replacement(self):
        if self.const_replacement:
            return False
        return ANTLRElement.calc_replacement(self)


class ANTLRLexerElements(ANTLRLexerElement):
    def calc_starters(self):
        if self.children[0].start_intervals and self.start_intervals is None:
            self.start_intervals = self.children[0].start_intervals
            return True
        return False


class ANTLRLexerAlternation(ANTLRLexerElement):
    def calc_replacement(self):
        # The replacement is the known shortest replacement of the children.
        if self.has_defined_replacement():
            new_repl = min(list(filter(lambda i: i is not None, [c.replacement for c in self.children])), key=len, default=False)
            if self.replacement is None or len(new_repl) < len(self.replacement) or (len(new_repl) == len(self.replacement) and new_repl < self.replacement):
                self.replacement = new_repl
                return True
        return False


class ANTLRTokenRef(ANTLRLexerElement):
    def __init__(self, ref):
        ANTLRLexerElement.__init__(self)
        self.ref = ref


class ANTLRCharacterRange(ANTLRLexerElement):
    def __init__(self, start, end):
        ANTLRLexerElement.__init__(self)
        # Converting unicode code points to integers.
        start = int(start.split('\\u')[1], 16) if start.startswith('\\u') else ord(start)
        end = int(end.split('\\u')[1], 16) if end.startswith('\\u') else ord(end)
        self.start_intervals = [(start, end)]
        self.replacement = chr(start)


class ANTLRDotElement(ANTLRLexerElement):
    def __init__(self, *, optional=False):
        ANTLRLexerElement.__init__(self, optional=optional)
        # Hard-wiring ASCII character range here does not have any limitation (neither effect).
        # Basically it should not be used anyway, since the replacement is
        # constantly set to 'a' and negating 'any character' would not make sense.
        self.start_intervals = [(0, 255)]
        if self.replacement is None:
            self.replacement = 'a'


class ANTLRString(ANTLRLexerElement):
    def __init__(self, src):
        ANTLRLexerElement.__init__(self)
        src = ANTLRSetElement.resolve_escapes(src)
        self.start_intervals = [(ord(src[0]), ord(src[0]))]
        self.replacement = src


class ANTLRSetElement(ANTLRLexerElement):
    def __init__(self, *, content=None, optional=False):
        ANTLRLexerElement.__init__(self, optional=optional)
        if content and self.replacement is None:
            if content.startswith(('"', '\'')):
                self.start_intervals = [(ord(content[1]), ord(content[1]))] if len(content) > 2 else []
                self.replacement = chr(self.start_intervals[0][0])
            elif content.startswith('['):
                self.start_intervals = ANTLRSetElement.process_charset(content[1:-1])
                self.replacement = chr(self.start_intervals[0][0])

    @staticmethod
    def process_charset(src):
        """
        Extract represented character intervals from chararcter sets.

        :param src: The string representation of the character set (w/o brackets).
        """
        intervals = [(ord(m.group(1)), ord(m.group(2))) for m in re.finditer('(\w)\-(\w)', src)]
        positions = [(m.start(1), m.end(2)) for m in re.finditer('(\w)\-(\w)', src)]
        return intervals + ANTLRSetElement.extract_single_chars(src, positions)

    @staticmethod
    def extract_single_chars(src, positions):
        """
        Character sets can contain multiple sets and single characters. (e.g. [-ab-defg-ijkl])
        This function selects the single characters based on the position of sets.

        :param src: The string representation of the character set (w/o brackets).
        :param positions: Position intervals in src where character intervals are placed.
        """
        if not positions:
            return [(ord(x), ord(x)) for x in list(ANTLRSetElement.resolve_escapes(src))]
        characters = []
        for i, pos in enumerate(positions):
            # Characters before the first range.
            if i == 0 and pos[0] > 0:
                characters.extend(list(ANTLRSetElement.resolve_escapes(src[0: pos[0]])))
            # Characters between ranges.
            if i < len(positions) - 1:
                if positions[i][1] + 1 < positions[i + 1][0]:
                    characters.extend(list(ANTLRSetElement.resolve_escapes(src[positions[i][1] + 1: positions[i + 1][0]])))
            # Characters after ranges.
            else:
                if pos[1] < len(src) - 1:
                    characters.extend(list(ANTLRSetElement.resolve_escapes(src[pos[1] + 1:])))
        return [(ord(x), ord(x)) for x in characters]

    @staticmethod
    def resolve_escapes(src):
        """
        Remove escaping from escape sequences in src. E.g. lexer rules may contain such
        expressions like: [\t] where \t is evaluated as '\' + 't' instead of a tabulator.
        This function executes the reversed transformation.

        :param src: The string that may have escaped escape sequences.
        """
        return bytes(src, 'utf-8').decode('unicode_escape')

    def calc_starters(self):
        if self.start_intervals is None and self.children and self.children[0].start_intervals:
            self.start_intervals = self.children[0].start_intervals
            return True
        return False


class ANTLRNotSet(ANTLRLexerElement):
    def calc_starters(self):
        # Known limitation (TODO?): it does not handle multiple negation.
        if self.starters_defined() and self.start_intervals is None:
            intervals = [y for x in self.children for y in x.start_intervals]
            # Sort list of tuples by the first element.
            intervals.sort(key=lambda x: x[0])
            # The number (char) before the first interval's lower limit is suitable for negation.
            self.start_intervals = [(intervals[0][0] - 1, intervals[0][0] - 1)]
            return True
        return False

    def calc_replacement(self):
        if self.start_intervals and self.replacement is None:
            self.replacement = chr(self.start_intervals[0][0])
            return True
        return False



# Copyright (c) 2016-2017 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

import antlerinator
import codecs
import json
import logging
import picire
import pkgutil

from argparse import ArgumentParser
from os.path import abspath, basename, exists, join, relpath
from shutil import rmtree

from antlr4 import *
from .antlr4 import create_hdd_tree, IslandDescriptor
from .coarse_hdd import coarse_hddmin, coarse_full_hddmin
from .hdd import hddmin

logger = logging.getLogger('picireny')
__version__ = pkgutil.get_data(__package__, 'VERSION').decode('ascii').strip()
antlr_default_path = antlerinator.antlr_jar_path


args_hdd_choices = {
    'full': hddmin,
    'coarse': coarse_hddmin,
    'coarse-full': coarse_full_hddmin,
}

def process_args(arg_parser, args):
    args.hddmin = args_hdd_choices[args.hdd]

    if args.antlr:
        if args.antlr == antlr_default_path:
            antlerinator.install(lazy=True)

        args.antlr = abspath(relpath(args.antlr))
        if not exists(args.antlr):
            arg_parser.error('%s does not exist.' % args.antlr)

    for i, g in enumerate(args.grammar):
        args.grammar[i] = abspath(relpath(g))
        if not exists(args.grammar[i]):
            arg_parser.error('%s does not exist.' % args.grammar[i])

    if args.replacements:
        if not exists(args.replacements):
            arg_parser.error('%s does not exist.' % args.replacements)
        else:
            try:
                with open(args.replacements, 'r') as f:
                    args.replacements = json.load(f)
            except json.JSONDecodeError as err:
                arg_parser.error('The content of %s is not a valid JSON object: %s' % err)
    else:
        args.replacements = {}

    if args.islands:
        if not exists(args.islands):
            arg_parser.error('%s does not exist.' % args.islands)
        with open(args.islands, 'r') as f:
            args.islands = json.load(f, object_hook=IslandDescriptor.json_load_object_hook)

    picire.cli.process_args(arg_parser, args)


def call(*,
         reduce_class, reduce_config,
         tester_class, tester_config,
         input, src, encoding, out,
         hddmin,
         antlr, grammar, start_rule, replacements=None, islands=None, lang='python',
         hdd_star=True, squeeze_tree=True, skip_unremovable_tokens=True,
         flatten_recursion=False,
         cache_class=None, cleanup=True):
    """
    Execute picireny as if invoked from command line, however, control its
    behaviour not via command line arguments but function parameters.

    :param reduce_class: Reference to the reducer class.
    :param reduce_config: Dictionary containing information to initialize the reduce_class.
    :param tester_class: Reference to a runnable class that can decide about the interestingness of a test case.
    :param tester_config: Dictionary containing information to initialize the tester_class.
    :param input: Path to the test case to reduce (only used to determine the name of the output file).
    :param src: Contents of the test case to reduce.
    :param encoding: Encoding of the input test case.
    :param out: Path to the output directory.
    :param hddmin: Function implementing a HDD minimization algorithm.
    :param antlr: Path to the ANTLR4 tool (Java jar binary).
    :param grammar: Path to the grammar(s) that can parse the top-level language.
    :param start_rule: Name of the start rule of the top-level grammar.
    :param replacements: Dictionary containing the minimal replacement of every lexer and parser rules.
    :param islands: Path to the Python3 file describing how to process island grammars.
    :param lang: The target language of the parser.
    :param hdd_star: Boolean to enable the HDD star algorithm.
    :param squeeze_tree: Boolean to enable the tree squeezing optimization.
    :param skip_unremovable_tokens: Boolean to enable hiding unremovable tokens from ddmin.
    :param flatten_recursion: Boolean to enable flattening left/right-recursive trees.
    :param cache_class: Reference to the cache class to use.
    :param cleanup: Binary flag denoting whether removing auxiliary files at the end is enabled (default: True).
    :return: The path to the minimal test case.
    """

    # Get the parameters in a dictionary so that they can be pretty-printed
    # (minus src, as that parameter can be arbitrarily large)
    args = locals().copy()
    del args['src']
    logger.info('Reduce session starts for %s\n%s',
                input, ''.join(['\t%s: %s\n' % (k, v) for k, v in sorted(args.items())]))

    grammar_workdir = join(out, 'grammar')
    tests_workdir = join(out, 'tests')

    hdd_tree = create_hdd_tree(InputStream(src.decode(encoding)), grammar, start_rule, antlr, grammar_workdir,
                               replacements=replacements, island_desc=islands, lang=lang)

    if flatten_recursion:
        hdd_tree.flatten_recursion()

    if squeeze_tree:
        hdd_tree = hdd_tree.squeeze_tree()

    if skip_unremovable_tokens:
        hdd_tree.skip_unremovable_tokens()

    # Start reduce and save result to a file named the same like the original.
    out_file = join(out, basename(input))
    with codecs.open(out_file, 'w', encoding=encoding, errors='ignore') as f:
        f.write(hddmin(hdd_tree,
                       reduce_class,
                       reduce_config,
                       tester_class,
                       tester_config,
                       basename(input),
                       tests_workdir,
                       hdd_star=hdd_star,
                       cache=cache_class() if cache_class else None))
    logger.info('Result is saved to %s.', out_file)

    if cleanup:
        rmtree(grammar_workdir)
        rmtree(tests_workdir)

    return out_file


def execute():
    """
    The main entry point of picireny.
    """

    arg_parser = ArgumentParser(description='CLI for the Picireny Hierarchical Delta Debugging Framework',
                                parents=[picire.cli.create_parser()], add_help=False)

    # Grammar specific settings.
    arg_parser.add_argument('--hdd', metavar='NAME', choices=args_hdd_choices.keys(), default='full',
                            help='HDD variant to run (%(choices)s; default: %(default)s)')
    arg_parser.add_argument('-s', '--start-rule', metavar='NAME', required=True,
                            help='start rule of the grammar')
    arg_parser.add_argument('-g', '--grammar', metavar='FILE', nargs='+', required=True,
                            help='grammar file(s) describing the input format')
    arg_parser.add_argument('-r', '--replacements', metavar='FILE',
                            help='JSON file defining the default replacements for lexer or parser rules')
    arg_parser.add_argument('--antlr', metavar='FILE', default=antlr_default_path,
                            help='path where the antlr jar file is installed (default: %(default)s)')
    arg_parser.add_argument('--islands', metavar='FILE',
                            help='JSON file describing how to process island languages')
    arg_parser.add_argument('--parser', metavar='LANG', default='python', choices=['python', 'java'],
                            help='language of the generated parsers (%(choices)s; default: %(default)s) '
                                 '(using Java might gain performance, but needs JDK)')
    arg_parser.add_argument('--no-hdd-star', dest='hdd_star', default=True, action='store_false',
                            help='run the hddmin algorithm only once')
    arg_parser.add_argument('--no-squeeze-tree', dest='squeeze_tree', default=True, action='store_false',
                            help='don\'t squeeze rule chains in tree representation')
    arg_parser.add_argument('--no-skip-unremovable-tokens', dest='skip_unremovable_tokens', default=True, action='store_false',
                            help='don\'t hide unremovable tokens from the ddmin algorithm')
    arg_parser.add_argument('--flatten-recursion', default=False, action='store_true',
                            help='flatten recurring blocks of left/right-recursive rules')
    arg_parser.add_argument('--version', action='version', version='%(prog)s {version}'.format(version=__version__))

    args = arg_parser.parse_args()
    process_args(arg_parser, args)

    logging.basicConfig(format='%(message)s')
    logger.setLevel(args.log_level)
    logging.getLogger('picire').setLevel(logger.level)

    call(reduce_class=args.reduce_class,
         reduce_config=args.reduce_config,
         tester_class=args.tester_class,
         tester_config=args.tester_config,
         input=args.input,
         src=args.src,
         encoding=args.encoding,
         out=args.out,
         hddmin=args.hddmin,
         antlr=args.antlr,
         grammar=args.grammar,
         start_rule=args.start_rule,
         lang=args.parser,
         replacements=args.replacements,
         islands=args.islands,
         hdd_star=args.hdd_star,
         squeeze_tree=args.squeeze_tree,
         skip_unremovable_tokens=args.skip_unremovable_tokens,
         flatten_recursion=args.flatten_recursion,
         cache_class=args.cache,
         cleanup=args.cleanup)



# Copyright (c) 2016 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

from antlr4 import *

from .antlr_tree import *


def analyze_grammars(antlr_lexer, antlr_parser, grammars, replacements):
    """
    Determine the minimal parser rule replacements of the input grammar.

    :param antlr_lexer: Reference to the ANTLR4 lexer class.
    :param antlr_parser: Reference to the ANTLR4 parser class.
    :param grammars: List of the grammars describing the input format.
    :param replacements: Dictionary that contains the predefined minimal
                         replacement of any lexer or parser rules. These won't be overridden later.
    :return: Pair of the replacement dictionary and the positions of quantified elements in the grammars.
    """

    def set_replacements(tree):
        """
        Set the minimal replacements of the various subtrees.

        :param tree: AST-like tree representation built by create_grammar_tree.
        """
        iterate = True
        # Iterate until any updates were performed.
        while iterate:
            iterate = False
            for e in tree:
                # If all of the children have a min set:
                s = isinstance(e, ANTLRLexerElement) and e.calc_starters()
                r = e.calc_replacement()
                if s or r:
                    iterate = True

    # Only those ParseTrees are present in our tree representation that
    # have real effect on the minimal replacements of the rules.
    # e.g. actions, channels, return values, syntax elements (like: |;:), etc
    # are avoided, but e.g. rule definitions, alternations, references,
    # token definitions or such nodes that can have quantifier are kept.
    def create_node(ctx, optional):
        """
        Create tree node of the lexer and parser subtrees.

        :param ctx: The ANTLRRuleContext object under processing.
        :param optional: Boolean indicating whether the current context/node is optional or not.
        :return: Node representation of the current context if needed, otherwise None.
        """

        # Parser rules.

        if isinstance(ctx, parser.ParserRuleSpecContext):
            # The parserRuleSpec rule contains 3 or 4 terminal tokens and one of them is the ID of the rule.
            # Since we cannot make a distinction between terminals at this point, they have to be referred
            # by indices. Since only the first terminal is optional indexing them from the back is safe
            # (the 3th from back is the rule ID).
            name = [x for x in ctx.children if isinstance(x, tree.Tree.TerminalNodeImpl)][-3].symbol.text
            return ANTLRRule(name, replacements.get(name, None))

        # Alternations need special handling since their minimal replacements are their shortest
        # child (in every other cases the children would be concatenated).
        if isinstance(ctx, (parser.AltListContext, parser.RuleAltListContext)):
            return ANTLRAlternation()

        # Node is created from Alternative to group its element+ children (it's a sequence).
        if isinstance(ctx, parser.AlternativeContext):
            return ANTLRElement(repl=('' if not ctx.children else None))

        # LabeledElement and Block are created since they can have quantifier.
        if isinstance(ctx, (parser.LabeledElementContext, parser.BlockContext)):
            return ANTLRElement(optional=optional)

        # Atom can also have quantifier. Furthermore it may have a terminal child
        # (DOT = matching any character) that has to be handled here.
        if isinstance(ctx, parser.AtomContext):
            if isinstance(ctx.children[0], tree.Tree.TerminalNodeImpl):
                assert ctx.children[0] == '.'
                return ANTLRDotElement(optional=optional)
            # Create a base ANTLRElement anyway to make possible applying the quantifier
            # to the subtree.
            return ANTLRElement(optional=optional)

        # Only the reference is set here but in the next step the whole referenced
        # subtree will be plugged as its child.
        if isinstance(ctx, parser.RulerefContext):
            assert ctx.getChildCount() == 1, 'RuleRef must have exactly one child.'
            return ANTLRRef(ctx.children[0].symbol.text, optional=optional)

        # Lexer rules.

        # The main difference between parser and lexer rules in this representation is that
        # lexer rules have an additional field (start_intervals) that aims to track all the
        # possible character ranges that a given token can start with. The purpose of this
        # is being able to generate minimal replacement for a negated lexer rule: having
        # all the possible character intervals that a lexer rule can start with we can easily
        # invert these ranges.
        if isinstance(ctx, parser.LexerRuleSpecContext):
            # Just like at ANTLRRule, the 3rd terminal from the back contains the name of the lexer rule.
            name = [x for x in ctx.children if isinstance(x, tree.Tree.TerminalNodeImpl)][-3].symbol.text
            return ANTLRLexerRule(name, repl=replacements.get(name, None))

        # The same logic as with parser alternations.
        if isinstance(ctx, parser.LexerAltListContext):
            return ANTLRLexerAlternation()

        # The special about LexerAlt is that it can have an empty child which makes
        # possible such alternations in lexer like: ('a'| ). Capturing an empty LexerAlt
        # construction is only possible here, in which case its minimal replacement is
        # the empty string.
        if isinstance(ctx, parser.LexerAltContext):
            # If the alternative has no children means that it's left explicitly empty.
            return ANTLRLexerElement(repl=('' if not ctx.children else None))

        # The special about LexerElements is that by determining its start character range
        # is enough to get the first character of its first child (since it's a token sequence).
        if isinstance(ctx, parser.LexerElementsContext):
            return ANTLRLexerElements()

        # LabeledLexerElement and LexerBlock are created since they can have quantifier.
        if isinstance(ctx, (parser.LabeledLexerElementContext, parser.LexerBlockContext)):
            return ANTLRLexerElement(optional=optional)

        # LexerAtom can also have quantifier. Furthermore it may have terminal children
        # (DOT or character set) that has to be handled here.
        if isinstance(ctx, parser.LexerAtomContext):
            if isinstance(ctx.children[0], tree.Tree.TerminalNodeImpl):
                content = ctx.children[0].symbol.text
                if content == '.':
                    return ANTLRDotElement(optional=optional)
                if content.startswith('['):
                    return ANTLRSetElement(content=content, optional=optional)
                assert False
            # Create a base ANTLRLexerElement anyway to make possible applying the
            # quantifier to the subtree.
            return ANTLRLexerElement(optional=optional)

        if isinstance(ctx, parser.CharacterRangeContext):
            # The 1st and 3rd token of a character range defines its boundaries.
            return ANTLRCharacterRange(ctx.children[0].symbol.text[1:-1], ctx.children[2].symbol.text[1:-1])

        if isinstance(ctx, parser.TerminalContext):
            # Terminal node is either a string literal or a token reference.
            content = ctx.children[0].symbol.text
            if content.startswith(('"', '\'')):
                return ANTLRString(content[1:-1])
            return ANTLRTokenRef(content)

        if isinstance(ctx, parser.NotSetContext):
            return ANTLRNotSet()

        # SetElement is the lexer rule that will be negated.
        if isinstance(ctx, parser.SetElementContext):
            # If the first child is a terminal node then it must be one of the followings:
            # token_ref, string_literal or char set.
            if isinstance(ctx.children[0], tree.Tree.TerminalNodeImpl):
                if ctx.children[0].symbol.text.isupper():
                    return ANTLRTokenRef(ctx.children[0].symbol.text)
                return ANTLRSetElement(content=ctx.children[0].symbol.text)
            # In this case we have a character range.
            return ANTLRSetElement()

        return None

    def get_quantifier(children, idx):
        """
        Check whether a quantifier is defined on the idx-th children.

        :param children: All the siblings of the current node.
        :param idx: The index of the current node among the siblings.
        :return: Quantifier string of the idx-th context if one is defined, None otherwise.
        """
        if len(children) <= idx + 1:
            return None
        suffix = None
        if isinstance(children[idx + 1], parser.EbnfSuffixContext):
            suffix = children[idx + 1].start.text
        elif isinstance(children[idx + 1], parser.BlockSuffixContext):
            suffix = children[idx + 1].children[0].start.text
        return suffix

    def is_optional(quantifier):
        """
        Check whether a quantifier string makes its quantified expression optional,
        i.e., if it allows the expression to occur 0 times.

        :param quantifier: Quantifier string.
        :return: Boolean indicating whether the quantifier is optional or not.
        """
        return quantifier.startswith(('*', '?'))

    def create_grammar_tree(node, positions, parent_idx, optional, parser_rule):
        """
        Creates a tree representation of the target parser grammar to facilitate
        the generation of minimal replacement strings.

        :param node: The ANTLR parser tree whose representation will be inserted now.
        :param positions: Dictionary describing positions in grammars where optional
                          actions should be injected.
        :param parent_idx: The index of the parent node in the elements list or None if without parent.
        :param optional: Boolean deciding if the current node is optional or not.
        :param parser_rule: Boolean value indicating if a parser rule being processed.
        """
        element = create_node(node, optional)
        if element:
            elements.append(element)
            idx = len(elements) - 1
            if parent_idx is not None:
                elements[parent_idx].children.append(element)
        else:
            idx = parent_idx

        if node.getChildCount() > 0:
            # TerminalNodeImpl nodes already have been added by create_node
            # when processing their parent since at this point we don't know their type.
            for i, c in enumerate([x for x in node.children if not isinstance(x, tree.Tree.TerminalNodeImpl)]):
                quantifier = get_quantifier(node.children, i)

                # Mark positions in parser rules that have any quantifier applied on them.
                if quantifier and parser_rule:
                    start_token = parser.getInputStream().get(c.getSourceInterval()[0])
                    end_token = parser.getInputStream().get(c.getSourceInterval()[1])

                    start_ln = start_token.line
                    start = start_token.column

                    line_breaks = end_token.text.count('\n')
                    end_ln = end_token.line + line_breaks
                    end = end_token.column + len(end_token.text) if not line_breaks else \
                        len(end_token.text) - end_token.text.rfind('\n') + 1

                    if start_ln not in positions:
                        positions[start_ln] = []
                    if end_ln not in positions:
                        positions[end_ln] = []

                    positions[start_ln].append(('s', start))
                    positions[end_ln].append(('e', end))

                create_grammar_tree(c, positions, idx, quantifier and is_optional(quantifier),
                                    parser_rule and not isinstance(element, ANTLRLexerRule))

    # EOF is a special token provided by the ANTLR framework. It's added preliminarily to
    # our tree to avoid dead links to it.
    elements = [ANTLRLexerRule('EOF', '')]
    action_positions = {}
    replacements = replacements if replacements else {}
    # Fill elements with node representations.
    for grammar in grammars:
        action_positions[grammar] = {}
        parser = antlr_parser(CommonTokenStream(antlr_lexer(FileStream(grammar, 'utf-8'))))
        create_grammar_tree(parser.grammarSpec(), action_positions[grammar], None, False, True)

    # Create mapping between references and indices of antlr_tree to be able to plug the
    # appropriate subtrees into reference nodes.
    rules = dict((x.name, i) for i, x in enumerate(elements) if isinstance(x, (ANTLRRule, ANTLRLexerRule)))

    # Plug the referred node under the referrers.
    for i, x in enumerate(elements):
        if isinstance(x, (ANTLRRef, ANTLRTokenRef)):
            assert len(elements[i].children) == 0, 'Referrer nodes must not contain children.'
            elements[i].children = [elements[rules[x.ref]]]

    # Associate tree nodes with minimal string replacements.
    set_replacements(elements)
    return dict((x.name, x.replacement) for x in elements if isinstance(x, (ANTLRRule, ANTLRLexerRule))), action_positions



# Copyright (c) 2017 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.


class Unparser(object):
    """Class defines how to build test case from an HDD tree."""

    def __init__(self, tree, ids):
        """
        Initialize the unparser object.

        :param tree: Tree representing the current test case.
        :param ids: The IDs of nodes that can change status.
        """
        self.tree = tree
        self.ids = ids

    def __call__(self, config):
        """
        :param config: List of IDs of nodes that will be kept in the next test case.
        :return: The unparsed test case containing only the units defined in config.
        """
        self.tree.set_state(self.ids, set(config))
        return self.tree.unparse()



from antlr4 import *


class LexerAdaptor(Lexer):

    """
    Track whether we are inside of a rule and whether it is lexical parser. _currentRuleType==Token.INVALID_TYPE
    means that we are outside of a rule. At the first sign of a rule name reference and _currentRuleType==invalid, we
    can assume that we are starting a parser rule. Similarly, seeing a token reference when not already in rule means
    starting a token rule. The terminating ';' of a rule, flips this back to invalid type.

    This is not perfect logic but works. For example, "grammar T;" means that we start and stop a lexical rule for
    the "T;". Dangerous but works.

    The whole point of this state information is to distinguish between [..arg actions..] and [charsets]. Char sets
    can only occur in lexical rules and arg actions cannot occur.
    """

    _currentRuleType = Token.INVALID_TYPE

    def __init__(self, inp):
        Lexer.__init__(self, inp)

    def getCurrentRuleType(self):
        return self._currentRuleType

    def setCurrentRuleType(self, ruleType):
        self._currentRuleType = ruleType

    def handleBeginArgument(self):
        if self.inLexerRule():
            self.pushMode(self.LexerCharSet)
            self.more()
        else:
            self.pushMode(self.Argument)

    def handleEndArgument(self):
        self.popMode()
        if len(self._modeStack) > 0:
            self._type = self.ARGUMENT_CONTENT

    def handleEndAction(self):
        self.popMode()
        if len(self._modeStack) > 0:
            self._type = self.ACTION_CONTENT

    def emit(self):
        if self._type == self.ID:
            firstChar = self._input.getText(self._tokenStartCharIndex, self._tokenStartCharIndex)
            if firstChar[0].isupper():
                self._type = self.TOKEN_REF
            else:
                self._type = self.RULE_REF

            if self._currentRuleType == Token.INVALID_TYPE:  # if outside of rule def
                self._currentRuleType = self._type  # set to inside lexer or parser rule

        elif self._type == self.SEMI:  # exit rule def
            self._currentRuleType = Token.INVALID_TYPE
        return Lexer.emit(self)

    def inLexerRule(self):
        return self._currentRuleType == self.TOKEN_REF

    def inParserRule(self):  # not used, but added for clarity
        return self._currentRuleType == self.RULE_REF



# Copyright (c) 2016-2017 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

from os.path import dirname, join
from setuptools import setup, find_packages

with open(join(dirname(__file__), 'picireny/VERSION'), 'rb') as f:
    version = f.read().decode('ascii').strip()


setup(
    name='picireny',
    version=version,
    packages=find_packages(),
    url='https://github.com/renatahodovan/picireny',
    license='BSD',
    author='Renata Hodovan, Akos Kiss',
    author_email='hodovan@inf.u-szeged.hu, akiss@inf.u-szeged.hu',
    description='Picireny Hierarchical Delta Debugging Framework',
    long_description=open('README.rst').read(),
    install_requires=['antlerinator==4.6', 'picire==17.6'],
    zip_safe=False,
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'picireny = picireny.cli:execute'
        ]
    },
)



# Copyright (c) 2016 Renata Hodovan, Akos Kiss.
#
# Licensed under the BSD 3-Clause License
# <LICENSE.rst or https://opensource.org/licenses/BSD-3-Clause>.
# This file may not be copied, modified, or distributed except
# according to those terms.

import re


class IslandDescriptor(object):
    """
    Describes island grammars within an input.

    The instructions to process the island grammars are defined by the user
    through a JSON dictionary.
    The dictionary must have 2 mandatory fields, and can optionally have 3
    additional fields:

    * rule: The name of the lexer rule in the host grammar containing the island.
    * grammars: List of grammars describing the island format.
    * pattern: Regular expression to be matched to the text of a HDDToken.
               Matching tokens will be replaced by subtrees. The matching
               named groups are parsed by the provided grammars using the
               group names as start rules. Intermediate substrings not matched
               by any group are represented by HDDTokens.
    * replacements: Dictionary containing pre-defined minimal replacements of the
                    target grammar's rules (if any).
    * island_desc: List of island descriptors for further islands within this
                   format. The tree can be arbitrarily deep.

    Example:
    {
        "rule": "STYLE_BODY",
        "grammars": ["/path/to/css/grammar.g4"],
        "pattern": "(?P<stylesheet>.*)</style>"
    }
    """

    def __init__(self, rule, grammars, *, pattern=None, replacements=None, island_desc=None):
        self.rule = rule
        self.grammars = grammars
        pattern = pattern if pattern else '(?P<{rule}>^.*$)'.format(rule=rule[0].lower() + rule[1:])
        self.pattern = re.compile(pattern, re.S)

        self.replacements = replacements if replacements is not None else {}
        self.island_desc = island_desc

    @staticmethod
    def json_load_object_hook(data):
        if 'rule' in data and 'grammars' in data:
            return IslandDescriptor(rule=data['rule'],
                                    grammars=data['grammars'],
                                    pattern=data.get('pattern', None),
                                    replacements=data.get('replacements', None),
                                    island_desc=data.get('island_desc', None))
        return data
